%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% File: parser.oz
% Purpose: parse an xml string to build Mozart representation
% Date: 2000 11 28
% Revised: 2000 12 13 ==>
%          added: empty element and cdata
%          2001 09 20 ==>
%          become namespace-aware
% Author: Mahmoud Rafea
% Copyright: Mahmoud Rafea
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%\define OPI
\ifdef OPI
declare
F =
\endif
functor
import
   Stack at 'stack.ozf'
export
   xmlDoc:XMLDoc
   xmlScan:XMLScan
   scanXML:ScanXML
   key:Key
   
define
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   %% Parse Processing Instruction but do nothing
   fun {PI S}
      case S of nil then raise xmlerror('Error in processing instruction') end
      [] &?|&>|Rest then Rest
      [] _|Cs then {PI Cs}
      end
   end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   %% Parse DTD string but do nothing
   fun {LoadDTD S Doc}
      case S of nil then raise xmlerror('Error in DTD instruction') end
      [] &>|Rest then Rest
      [] _|Cs then {LoadDTD Cs Doc}
      end
   end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   %% Global flags
   %% Becomes true when parsing a node tag similar to <n />
   Empty = {NewCell false}
   %% Becomes true when node has attributes
   Attr = {NewCell false}
   %% Becomes true when a node has an explicit name space
   NameSpace = {NewCell false}
   %%
   TopNode = {NewCell true}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   %% Removes starting double or single quote
   fun {Refine Cs}
      case Cs
      of nil then nil
      [] &"|Xs then Xs
      [] &'|Xs then Xs
      else Cs
      end
   end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   %% Return with a list of node attributes.
   %% The attribute is tuple of the following form:
   %%     name#value#namespace#basename
   %% name = basename if namespace = ''
   %% if a default namesapce is define, it is pushed in the namespace-stack
   proc {CollectAttr S Doc NSSt As Rest}
      fun {GetValue S S1}
	 case S
	 of nil then S1 = nil nil
	 [] &"|Cs then S1 = Cs nil
	 [] &'|Cs then S1 = Cs nil
	 [] C|Cs then C|{GetValue Cs S1}
	 end
      end
      fun {GetName S S1}
	 case S
	 of nil then S1 = nil nil
	 [] C|Cs andthen {Char.isSpace C} then {String.token Cs &= _ S1} nil
	 [] &=|Cs then S1 = Cs nil
	 [] C|Cs then C|{GetName Cs S1}
	 end
      end
      fun {GetAttr S S1}
	 S2 NS Attr NSV AttrV
	 AttrName = {GetName S S2}
	 Value = {StringToAtom {GetValue {Refine S2} S1}}
      in
	 {String.token AttrName &: NS Attr}
	 case Attr of "" then AttrV = {StringToAtom NS}
	    if AttrV == xmlns then
	       AName = {NewName} in
	       NSV = 'w3c'
	       {Doc get(ns $)}.AName := Value
	       {Stack.push NSSt AName}
	    else NSV = '' end
	 else NSV = {StringToAtom NS} AttrV = {StringToAtom Attr}
	    if NSV == xmlns then
	       {Doc get(ns $)}.AttrV := Value
	    end
	 end
	 {StringToAtom AttrName}#Value#NSV#AttrV
      end
   in
      case S
      of nil then
	 raise xmlerror('file is corrupted') end
      [] &>|Cs then As = nil Rest = Cs
      [] &/|&>|Cs then {Assign Empty true} As = nil Rest = Cs
      [] C|Cs andthen {Char.isSpace C} then {CollectAttr Cs Doc NSSt As Rest}
      else
	 As1 S1 A = {GetAttr S S1} in
	 As = A|As1
	 {CollectAttr S1 Doc NSSt As1 Rest}
      end
   end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   %% Return with the node tag name
   %% sets the empty flag, if node is close
   %% set namespace flag, if namespace is explicity declared
   %% set attr flag, if node has attributes
   fun {TagName S S1} 
      case S
      of nil then raise xmlerror('unclosed tag') end
      [] &>|L then S1 = L nil
      [] &/|&>|L then {Assign Empty true} S1 = L nil
      [] 32|L then {Assign Attr true} S1 = L nil
      [] &:|L then {Assign NameSpace true} S1 = L nil
      [] X|L then X|{TagName L S1}
      end
   end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   %% The key of a node in the dictionary is an atom that is defined based on
   %% its position on the xml tree. Top node key is '0'. First child is '00'.
   %% Second child is '01'. First child of second child of top node is '010'.
   %% etc.
   fun {Key}
      {NewName}
   end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   %% Parse a new node
   fun {OpenTag S Doc}
      %% Returns the default namespace, if defined. Otherwise returns ''
      fun {CheckForNsInside Attrs}
	 case Attrs of H|Rest then
	    case H of xmlns#_#''#_ then
	       if {Stack.isEmpty NSSt} then raise xmlerror('Name space stack error') end
	       else {Stack.top NSSt} end
	    else {CheckForNsInside Rest}
	    end
	 else
	    ''
	 end
      end
      {Assign Empty false}
      {Assign Attr false}
      {Assign NameSpace false}
      Tag TName NS Rest Rest1 Rest2 P K As NSFlag
      St = {Doc get(s $)}
      NSSt = {Doc get(nss $)}
      Tag1 = {TagName S Rest2}
      N = {Doc get(n $)}+1
   in
      {Doc set(n N)}
      if {Access NameSpace} then
	 NS = {StringToAtom Tag1}
	 TName = {StringToAtom {TagName Rest2 Rest1}}
	 Tag = {VirtualString.toAtom NS#':'#TName}
      else
	 TName = {StringToAtom Tag1} Tag = {StringToAtom Tag1} Rest1 = Rest2
      end
      if {Stack.isEmpty St} then
	 P = 0
	 K = '00'
      else
	 P = {Stack.top St}.2
	 K = {NewName}
      end
      if {Access Attr} then 
	 {CollectAttr Rest1 Doc NSSt As Rest}
	 {Dictionary.put {Doc get(nattrs $)} K r(key:K nattrs:As)}
      else
	 As = nil
	 Rest = Rest1
      end
      %% Define a flag to know if default namespace exist
      %% Assign the correct namespace
      if {Access NameSpace} then NSFlag = false
      else
	 NS1 = {CheckForNsInside As} in
	 if NS1 == '' then
	    if {Stack.isEmpty NSSt} then NS = ''
	    else NS = {Stack.top NSSt} end
	    NSFlag = false
	 else NSFlag = true NS = NS1
	 end
      end
      if {Access Empty} then
	 if NSFlag then _ = {Stack.pop NSSt} end
      else
	 {Stack.push St Tag#K#NSFlag}
      end
      {Dictionary.put {Doc get(nodes $)} K node(key:K name:Tag tagname:TName namespace:NS parent:P n:N)}
      Rest
   end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   %% Vrifies that a node is closed.
   %% Pops up statcks and increments the tree level
   fun {CloseTag S Doc}
      {Assign NameSpace false}
      Rest Rest1 Tag
      Tag1 = {TagName S Rest1}
      St = {Doc get(s $)}
      NSSt = {Doc get(nss $)}
   in
      if {Access NameSpace} then
	 TName = {TagName Rest1 Rest} in
	 Tag = {VirtualString.toAtom Tag1#':'#TName}
      else
	 Tag = {StringToAtom Tag1} Rest = Rest1
      end
      if {Stack.isEmpty St} then
	 raise xmlerror('Unneeded close tag: ' {StringToAtom Tag}) end
      else Tuple NSFlag in
	 try
	    Tuple = {Stack.pop St}
	    Tag#_#NSFlag = Tuple
	    if NSFlag then _ = {Stack.pop NSSt} end
	    Rest
	 catch _ then
	    raise xmlerror('No matching tag: ' r(tag:{StringToAtom Tag} stack:Tuple)) end
	 end
      end
   end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   %% Scanning is transforming special xml chars into their escape sequence.
   %% It is done for node text
   %% XMLScan is the reverse of ScanXML(see below)
   local
      proc {Token S S1 T}
	 case S of nil then S1 = nil T = nil
	 [] C|C1|Cs then
	    case C
	    of &< then S1 = C1|Cs T = "&lt;"
	    [] &> then S1 = C1|Cs T = "&gt;"
	    [] && then S1 = C1|Cs T = "&amp;"
	       %% For compatability with Microsoft XML parser the apostrophe
	       %% and quot are not escaped
%	    [] &' then S1 = C1|Cs T = "&apos;"
%	    [] &" then S1 = C1|Cs T = "&quot;"
	    else
	       AN1 = {Char.isAlNum C}	    Sp1 = {Char.isSpace C}
	       AN2 = {Char.isAlNum C1}	    Sp2 = {Char.isSpace C1} in
	       if AN1 == AN2 orelse Sp1 == Sp2 then
		  S2 T1 in S1 = S2 T = C|T1 {Token C1|Cs S2 T1}
	       else
		  S1 = C1|Cs T = [C]
	       end
	    end
	 [] [C] then S1 = nil
	    case C
	    of &< then T = "&lt;"
	    [] &> then T = "&gt;"
	    [] && then T = "&amp;"
	    else T = [C] end
	 else
	    raise xmlerror({StringToAtom S}) end	 
	 end
      end
   in
      fun {XMLScan S}
	 S1 T = {Token S S1} in
	 if S1 == nil then T
	 elseif T == "" then
	    {VirtualString.toString {XMLScan S1}}
	 else
	    {VirtualString.toString T#{VirtualString.toString {XMLScan S1}}}
	 end
      end
   end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%   
   %% Scanning is transforming special xml chars escape sequence into chars.
   %% It is done for node text
   %% ScanXML is the reverse of XMLScan
   local
      proc {XMLToken S S1 T}
	 case S of nil then S1 = nil T = nil
	 [] &&|Rest then
	    case Rest of &a|&m|&p|&;|R then S1 = R T = "&amp;"
	    [] &l|&t|&;|R then S1 = R T = "&lt;"
	    [] &g|&t|&;|R then S1 = R T = "&gt;"
	    [] &a|&p|&o|&s|&;|R then S1 = R T = "&apos;"
	    [] &q|&u|&o|&t|&;|R then S1 = R T = "&quot;"
	    else raise xmlerror('Bad escape sequence' {StringToAtom Rest}) end
	    end
	 [] C|&&|Rest then S1 = &&|Rest T = [C]
	 [] C|C1|Rest then
	    AN1 = {Char.isAlNum C}	 Sp1 = {Char.isSpace C}
	    AN2 = {Char.isAlNum C1} Sp2 = {Char.isSpace C1} in
	    if AN1 == AN2 orelse Sp1 == Sp2 then
	       S2 T1 in S1 = S2 T = C|T1 {XMLToken C1|Rest S2 T1}
	    else S1 = C1|Rest T = [C] end
	 [] [C] then S1 = nil T = [C]
	 else raise xmlerror({StringToAtom S}) end	end
      end
   in
      fun {ScanXML S}
	 S1 
	 T = {XMLToken S S1}
      in
	 if S1 == nil then
	    try Tokens2XML.{StringToAtom T}  catch _ then T end
	 elseif T == "" then
	    {VirtualString.toString {ScanXML S1}}
	 else
	    try
	       T1 = Tokens2XML.{StringToAtom T} in
	       {VirtualString.toString T1#{VirtualString.toString {ScanXML S1}}}
	    catch _ then
	       {VirtualString.toString T#{VirtualString.toString {ScanXML S1}}}
	    end
	 end
      end
   end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%   
   %% Parse node data
   fun {ReadData S Doc}
      fun {GetData S1 S2}
	 case S1
	 of nil then raise xmlerror('No matching tag') end
	 [] &<|&/|Rest then S2 = &<|&/|Rest nil
	 [] &<|Rest then S2 = &<|Rest nil
	 [] X|Rest then X|{GetData Rest S2}
	 end
      end
      Data1 = {GetData S S2}
      St = {Doc get(s $)}
      K S2 Data
   in
      if {Stack.isEmpty St} then K = '0'
      else K = {Stack.top St}.2 end
      Data = {ScanXML Data1}
      if Data == "" then skip
      else {Dictionary.put {Doc get(data $)} K data(key:K data:Data)} end
      S2
   end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   %% Get comments and store them
   fun {StoreComment S Doc}
      fun {GetComment S S1}
	 case S of nil then raise xmlerror('No matching tag') end
	 [] &-|&-|&>|Rest then S1 = Rest nil
	 [] C|Cs then C|{GetComment Cs S1}
	 end
      end
      S1 Comment = {GetComment S S1}
      St = {Doc get(s $)}
      K
   in
      if {Stack.isEmpty St} then K = '0'
      else K = {Stack.top St}.2 end
      {Dictionary.put {Doc get(comments $)} K r(key:K comment:Comment)}
      S1
   end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   %% Get CData and store it
   fun {StoreCData S Doc}
      fun {GetCData S S1}
	 case S of nil then raise xmlerror('No matching tag') end
	 [] &]|&]|&>|Rest then S1 = Rest nil
	 [] C|Cs then C|{GetCData Cs S1}
	 end
      end
      S1 CData = {GetCData S S1}
      St = {Doc get(s $)}
      K
   in
      if {Stack.isEmpty St} then K = '0'
      else K = {Stack.top St}.2 end
      {Dictionary.put {Doc get(cdata $)} K r(key:K cdata:CData)}
      S1
   end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   %% Praser main
   proc {Parse S Doc}
      S1 in
      case S
      of nil                             then skip
      [] &<|&?|Rest                      then S1 = {PI Rest}               {Parse S1 Doc}
      [] &<|&!|&D|&O|&C|&T|&Y|&P|&E|Rest then S1 = {LoadDTD Rest Doc}      {Parse S1 Doc}
      [] &<|&!|&-|&-|Rest                then S1 = {StoreComment Rest Doc} {Parse S1 Doc}
      [] &<|&!|&[|&C|&D|&A|&T|&A|&[|Rest then S1 = {StoreCData Rest Doc}   {Parse S1 Doc}
      [] &<|&/|Rest                      then S1 = {CloseTag Rest Doc}	   {Parse S1 Doc}
      [] &<|Rest                         then S1 = {OpenTag Rest Doc} {Assign TopNode false} {Parse S1 Doc}
      [] X|Rest andthen {Char.isSpace X} then {Parse Rest Doc}
      else S1 = {ReadData S Doc} {Parse S1 Doc}
      end
   end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   %%
   class XMLDoc from BaseObject
      attr
	 %% Mozart XML representation. Better names for dictionaries are needed.
	 nodes                     % dictionary of nodes
	 nattrs                    % dictionary of nodes attributes
	 ns                        % dictionary of namespaces
	 data                      % dictionary of nodes data
	 comments                  % dictionary of comments
	 cdata                     % dictionary of cdata
	 %% Stacks supporting parsing
	 s
	 nss
	 %% Tree level, which supports making a key for nodes in the dictionary
	 n
	 %%
      meth init
	 nodes <- {Dictionary.new}
	 nattrs <- {Dictionary.new}
	 ns <- {Dictionary.new}
	 data <- {Dictionary.new}
	 comments <- {Dictionary.new}
	 cdata <- {Dictionary.new}
	 s <- {Stack.new}
	 nss <- {Stack.new}
	 n <- 0
      end
      %%
      meth loadXML(S)
	 {Parse S self}
      end
      %%
      meth getChildren(Parent Nodes)
	 Nodes = {Filter {Dictionary.items @nodes}
		  fun {$ I} I.parent == Parent end}
      end
      %%
      meth get(A V)
	 V = @A
      end
      %%
      meth set(A V)
	 A <- V
      end
      %%
   end

   Tokens2XML = tokens('&amp;':"&" '&lt;':"<" '&gt;':">")
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% '&apos;':'\'' '&quot;':'"')
   
\ifdef OPI
[M] = {Module.apply [F]}

O = {New M.xmlDoc init}

{O loadXML("<a id=\"1\">data<b id='2'/></a><!-- Aladdin -->")}

{Browse {Dictionary.entries {O get(comments $)}}}
\endif